/** @file main.cpp
 * @brief  PhysX SDK̊b̊bBE{bNX쐬ėB
 * @author Yutaka Yoshisaka
 * @date   2006.09.08 - 2006.09.20
 */

#define NOMINMAX			// #include <windows.h>̑OɕKv
#include <windows.h>
#include <windowsx.h>

#include "mmsystem.h"

#include <GL/glut.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "NxPhysics.h"
#include "ErrorStream.h"

#ifdef _MSC_VER
#pragma warning(disable:4311 4312)  // LXg̃TCYϊwarning̖
#endif

#define OPENGL_DRAW_FLAG		1		///< OpenGL̕`sꍇ1w

#define MAX_ACTOR_COUNT		10000		///< zuAN^[̍ő吔

// [UƂĕێ邽߂̌`̎
#define SHAPE_TYPE_PLANE		1	///< ʂ̏ꍇ
#define SHAPE_TYPE_SPHERE		2	///< ̏ꍇ
#define SHAPE_TYPE_BOX			3	///< {bNX̏ꍇ

// 쐬鋅̃TCY
#define SHAPE_SPHERE_RADIUS		4.0f

// 쐬{bNX̃TCY
#define SHAPE_BOX_SIZE			8.0f

//----------------------------------------------------------//
// PhysXł̕ZɕKvȃO[oϐ					//
//----------------------------------------------------------//
static NxPhysicsSDK *g_pPhysicsSDK = NULL;		///< PhysX̃RANX
static NxScene      *g_pScene      = NULL;		///< PhysX̃V[NX
static int g_sphereCount;						///< AN^[ƂĂ̋̐
static int g_boxCount;							///< AN^[ƂẴ{bNX̐

//----------------------------------------------------------//
//	֐vg^Cv錾									//
//----------------------------------------------------------//
bool InitPhysX();
bool TermPhysX();
bool IsPhysXHardwarePresent();

bool AppendSphere(const NxVec3 &pos, const float r);		// zu
bool AppendBox(const NxVec3 &pos, const float size);		// {bNXzu

/**
 * 𕨗Ԃɔzu
 * @param[in] pos ԏ̈ʒu
 * @param[in] r   ̔a
 * @return        ɐtrueԂB
 */
bool AppendSphere(const NxVec3 &pos, const float r)
{
	NxActorDesc actorDesc;
	NxVec3 v;
	NxBodyDesc bodyDesc;
	NxSphereShapeDesc sphereDesc;
	NxActor *pActor;

	actorDesc.setToDefault();
	bodyDesc.setToDefault();
	sphereDesc.setToDefault();

	bodyDesc.angularDamping	= 0.5f;

	sphereDesc.radius = r;					// ƂĔaw(m)
	actorDesc.shapes.pushBack(&sphereDesc);
	actorDesc.body			= &bodyDesc;
	actorDesc.density		= 0.01f;		// ʖx(kg / m^3)
	actorDesc.globalPose.t  = pos;			// [hWʒu
	pActor = g_pScene->createActor(actorDesc);
	if(!pActor) {
		return false;
	}

	// [Uf[^̕ێ
	pActor->userData = (void *)(SHAPE_TYPE_SPHERE);

	return true;
}

/**
 * {bNX𕨗Ԃɔzu
 * @param[in] pos  ԏ̈ʒu
 * @param[in] size {bNX̃TCY
 * @return         ɐtrueԂB
 */
bool AppendBox(const NxVec3 &pos, const float size)
{
	NxActorDesc actorDesc;
	NxVec3 v;
	NxBodyDesc bodyDesc;
	NxBoxShapeDesc boxDesc;
	NxActor *pActor;

	actorDesc.setToDefault();
	bodyDesc.setToDefault();
	boxDesc.setToDefault();

	bodyDesc.angularDamping	= 0.5f;

	boxDesc.dimensions = NxVec3((float)size, (float)size, (float)size);
	actorDesc.shapes.pushBack(&boxDesc);
	actorDesc.body			= &bodyDesc;
	actorDesc.density		= 0.01f;		// ʖx(kg / m^3)
	actorDesc.globalPose.t  = pos;			// [hWʒu
	pActor = g_pScene->createActor(actorDesc);
	if(!pActor) {
		return false;
	}

	// [Uf[^̕ێ
	pActor->userData = (void *)(SHAPE_TYPE_BOX);

	return true;
}


/**
 * PhysX̃n[hEFAgp\ǂ
 * @return n[hEFAT|[gĂꍇtrueԂB
 */
bool IsPhysXHardwarePresent()
{
	if(!g_pPhysicsSDK) return false;
	return (g_pPhysicsSDK->getHWVersion() != NX_HW_VERSION_NONE);
}

/**
 * PhysX̏
 * @return ̐(true)Es(false)
 */
bool InitPhysX()
{
	int loop;
	NxVec3 v;

	// PhysX SDK̏
	g_pPhysicsSDK = NxCreatePhysicsSDK(NX_PHYSICS_SDK_VERSION, NULL, new ErrorStream());
	if(g_pPhysicsSDK == NULL) return false;

	g_pPhysicsSDK->setParameter(NX_SKIN_WIDTH, 0.05f);	// ڐGɗLɂ[

	// V[쐬
	NxSceneDesc sceneDesc;
	sceneDesc.gravity = NxVec3(0.0f, -9.8f, 0.0f);		// d͉͂
	
	if(IsPhysXHardwarePresent()) {							// n[hEFALǂ
		sceneDesc.simType = NX_SIMULATION_HW;				// PhysXn[hEFALɂ
		sceneDesc.hwSceneType = NX_HW_SCENE_TYPE_RB;		// ̕Ńn[hEFAgp
	}

	g_pScene = g_pPhysicsSDK->createScene(sceneDesc);
	if(g_pScene == NULL) return false;

	// ftHg̃}eAV[ɐݒ
	NxMaterial *defaultMaterial;
	defaultMaterial = g_pScene->getMaterialFromIndex(0);
	defaultMaterial->setRestitution(0.2f);				// ˕ԂW̎w
	defaultMaterial->setStaticFriction(0.5f);			// Î~CW̎w
	defaultMaterial->setDynamicFriction(0.5f);			// CW̎w

	// nʂzu
	NxPlaneShapeDesc planeDesc;							// nʂ̃IuWFNgiftHgȂ̂ Y = 0ɕʂłj
	planeDesc.normal = NxVec3(0.0f, 1.0f, 0.0f);
	planeDesc.d      = 0.0f;
	NxActorDesc actorDesc;
	actorDesc.shapes.pushBack(&planeDesc);				// AN^[ƂĒnʂvbV
	g_pScene->createActor(actorDesc);					// V[ɑ΂Ďw̃AN^[쐬

	// Ԃ̕ix߂邽߂ɕKvj
	g_pScene->setTiming((float)(1.0 / 60.0), 20);

	// ܂̓{bNXzu
	g_sphereCount = 0;
	g_boxCount = 0;
	for(loop = 0; loop < MAX_ACTOR_COUNT; loop++) {

		v.x = (float)(rand() & 255) - 128;
		v.y = (float)(rand() & 255) + 100.0f;
		v.z = (float)(rand() & 255) - 128;

		if(rand() & 1) {
			// AN^[Ƃč쐬
			if(!AppendSphere(v, SHAPE_SPHERE_RADIUS)) break;
			g_sphereCount++;
		} else {
			// {bNXAN^[Ƃč쐬
			if(!AppendBox(v, SHAPE_BOX_SIZE)) break;
			g_boxCount++;
		}
	}

	return true;
}

/**
 * PhysX̔j
 * @return trueԂB
 */
bool TermPhysX()
{
	// V[NXPhysX SDKNX
	if(g_pPhysicsSDK) {
		if(g_pScene) {
			g_pPhysicsSDK->releaseScene(*g_pScene);
		}
		g_pPhysicsSDK->release();

		g_pScene      = NULL;
		g_pPhysicsSDK = NULL;
	}

	return true;
}

/**
 * GLUT̃_OR[obN
 */
static void RenderCallback()
{
	char szStr[256];
	float glMat[16];
	long stTim, endTim;
	float dt;

	if(g_pScene == NULL) return;

	stTim = timeGetTime();

#if OPENGL_DRAW_FLAG
	// wĩNA
	glClearColor(0.3f, 0.4f, 0.5f, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// es̎w
	NxVec3 EyePos = NxVec3(0.0f, 0.0f, 0.0f);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0f, (float)glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), 1.0f, 10000.0f);
	gluLookAt(EyePos.x - 50.0f, EyePos.y + 100.0f, EyePos.z + 300.0f, EyePos.x, EyePos.y, EyePos.z, 0.0f, 1.0f, 0.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
#endif

	// Ԃ̎Ԃi߂
	dt = 1.0f / 60.0f;
	g_pScene->simulate(dt);

	//--------------------------------------------------//
	// AN^[̃_O							//
	//--------------------------------------------------//
	int actorCount, loop;
	NxActor **ppActors;
	NxActor *pActor;
	bool isDynamic;
	int userData;

	// AN^[̐擾
	actorCount = g_pScene->getNbActors();
	ppActors = g_pScene->getActors();

	for(loop = 0; loop < actorCount; loop++) {
		pActor = *ppActors;

		isDynamic = pActor->isDynamic();		// IȃIuWFNg̏ꍇ

		userData = (int)(pActor->userData);		// [Uf[^̎oił͌`̎ނĂj

#if OPENGL_DRAW_FLAG
		if(isDynamic) {

			if(userData == SHAPE_TYPE_SPHERE) {
				// ̏ꍇ
				glPushMatrix();
				pActor->getGlobalPose().getColumnMajor44(glMat);
				glMultMatrixf(glMat);
				glColor4f(1.0f, 0.0f, 0.0f, 1.0f);

				glutSolidSphere(SHAPE_SPHERE_RADIUS, 10, 10);		// ̔aƕw肵ċ`
				glPopMatrix();
			} else if(userData == SHAPE_TYPE_BOX) {
				// {bNX̏ꍇ
				glPushMatrix();
				pActor->getGlobalPose().getColumnMajor44(glMat);
				glMultMatrixf(glMat);
				glColor4f(1.0f, 0.0f, 0.0f, 1.0f);

				glutSolidCube(SHAPE_BOX_SIZE);						// {bNX̃TCYw肵ĕ`
				glPopMatrix();
			}

		} else {
			// nʂ̏ꍇisjɁAʂ`
			glPushMatrix();
			pActor->getGlobalPose().getColumnMajor44(glMat);
			glMultMatrixf(glMat);
			glColor4f(1.0f, 0.7f, 0.6f, 1.0f);

			glBegin(GL_QUADS);
			glNormal3f(0.0f, 1.0f, 0.0f);
			glVertex3f(-500.0f, 0.0f, 500.0f);
			glNormal3f(0.0f, 1.0f, 0.0f);
			glVertex3f(-500.0f, 0.0f, -500.0f);
			glNormal3f(0.0f, 1.0f, 0.0f);
			glVertex3f(500.0f, 0.0f, -500.0f);
			glNormal3f(0.0f, 1.0f, 0.0f);
			glVertex3f(500.0f, 0.0f, 500.0f);
			glEnd();

			glPopMatrix();
		}
#endif
		ppActors++;
	}

	// V[̗vffoĕύXLɂ
	g_pScene->flushStream();
	g_pScene->fetchResults(NX_RIGID_BODY_FINISHED, true);

	// OpenGL̃XN[XV
#if OPENGL_DRAW_FLAG
	glutSwapBuffers();
#endif

	endTim = timeGetTime();
	sprintf(szStr, "%d ̋ / %d ̃{bNX̕vZ`掞 = %d ms\n", g_sphereCount, g_boxCount, endTim - stTim);
	printf(szStr);
}

/**
 * GLUT̃r[|[gTCYύX̃R[obN
 */
static void ReshapeCallback(int width, int height)
{
	glViewport(0, 0, width, height);
}

/**
 * GLUT̃ACh̃R[obN
 */
static void IdleCallback()
{
	glutPostRedisplay();
}


//**************************************************************************//
/**
 * main֐
 */
int main(int argc, char *argv[])
{
	//--------------------------------------------------//
	// GLUT̏									//
	//--------------------------------------------------//
	glutInit(&argc, argv);
	glutInitWindowSize(512, 512);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	int mainHandle = glutCreateWindow("PhysXeXg");
	glutSetWindow(mainHandle);
	glutDisplayFunc(RenderCallback);
	glutReshapeFunc(ReshapeCallback);
	glutIdleFunc(IdleCallback);

	// OpenGL̃p[^w
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);

	// zuB͌Œ̕sƂ
	GLfloat light_amb[]  = {0.1f, 0.1f, 0.1f, 1.0f};
	GLfloat light_diff[] = {1.0f, 1.0f, 1.0f, 1.0f};
	GLfloat light_spec[] = {1.0f, 1.0f, 1.0f, 1.0f};
	GLfloat light_pos[]  = {0.0f, 30.0f, 20.0f, 0.0f};		// w = 0.0ŕsɂȂ

	glEnable(GL_LIGHTING);		// s
	glEnable(GL_LIGHT0);		// Cg0L
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_amb);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diff);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_spec);
	glLightfv(GL_LIGHT0, GL_POSITION, light_pos);

	// }eAw
	GLfloat material_diffuse[4]   = { 1.0f, 0.0f, 0.0f, 1.0f };
	GLfloat material_specular[4]  = { 1.0f, 1.0f, 1.0f, 1.0f };
	GLfloat material_emission[4]  = { 0.0f, 0.0f, 0.0f, 1.0f };
	GLfloat material_shininess[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
	material_shininess[0] = 100.0f;

	glMaterialfv(GL_FRONT, GL_DIFFUSE,   material_diffuse );
	glMaterialfv(GL_FRONT, GL_SPECULAR,  material_specular );
	glMaterialfv(GL_FRONT, GL_SHININESS, material_shininess );
	glMaterialfv(GL_FRONT, GL_EMISSION,  material_emission );

	//--------------------------------------------------//
	// PhysX̏									//
	//--------------------------------------------------//
	if(!InitPhysX()) return 0;

	//--------------------------------------------------//
	// GLUT̃C[v								//
	//--------------------------------------------------//
	glutMainLoop();

	//--------------------------------------------------//
	// PhysX̔j									//
	//--------------------------------------------------//
	TermPhysX();

	return 0;
}

